local TextSubs = FlexBar2:NewModule("TextSubs", "AceEvent-2.0");
-- If this var is incremented, SetDefaults() will be called for this module thus resetting all its settings.
TextSubs.Version = 1;

local TextSubList = {};
TextSubs.TextSubList = TextSubList;

-- Register a new sub
function TextSubs:RegisterTextSub(Textsub, Replace)
    if((type(Replace) == "function" or type(Replace) == "string") and type(Textsub) == "string") then
		FlexBar2:Debug("Registering new textsub: ", Textsub);
        TextSubList[Textsub] = Replace;
	else
		FlexBar2:Debug("failure to register ", Textsub, " as new textsub");
    end
    TextSubs:TriggerEvent("FB2_NewTextSub");
end

function TextSubs:UpdateTextSub(TextSub, ...)
	-- Loop through all buttons and call UpdateTextSub on them with the same arguments as here
	for ButtonName, ButtonRef in pairs(FlexBar2.ButtonList) do
		ButtonRef:UpdateTextSub(TextSub, ...);
	end
end

-- Declare the button mixin
TextSubs.ButtonMixin = FlexBar2.OO.Mixin
{
    "SetTextField",
	"UpdateTextSub",
	"MatchTextField",
};

function TextSubs.ButtonMixin:Initialize()
	-- Textsub cache of currently set subs
    self.TextSubs = {};
    for Key = 1, 5 do
        self.TextSubs[Key] = {SubCache = {}, PrefixList = {}, SuffixList = {}, ParsedText = ""};
    end
	
	local TextFields = {};
	self.TextFields = TextFields;
	-- Declare the first text field at the TOPLEFT
    local Text1 = self.Frame:CreateFontString(nil, "ARTWORK");
    Text1:SetFontObject("NumberFontNormal");
    Text1:SetJustifyH("LEFT");
    Text1:SetPoint("TOPLEFT", self.Frame, "TOPLEFT");
    self.TextFields[1] = Text1;
    
	-- Declare the second text field at the TOPRIGHT
    local Text2 = self.Frame:CreateFontString(nil, "ARTWORK");
    Text2:SetFontObject("NumberFontNormal");
    Text2:SetJustifyH("RIGHT");
    Text2:SetPoint("TOPRIGHT", self.Frame, "TOPRIGHT");
    self.TextFields[2] = Text2;
    
	-- Declare the third text field at the BOTTOMRIGHT
    local Text3 = self.Frame:CreateFontString(nil, "ARTWORK");
    Text3:SetFontObject("NumberFontNormal");
    Text3:SetJustifyH("RIGHT");
    Text3:SetPoint("BOTTOMRIGHT", self.Frame, "BOTTOMRIGHT");
    self.TextFields[3] = Text3;

	-- Declare the fourth text field at BOTTOMLEFT
    local Text4 = self.Frame:CreateFontString(nil, "ARTWORK");
    Text4:SetFontObject("NumberFontNormal");
    Text4:SetJustifyH("LEFT");
    Text4:SetPoint("BOTTOMLEFT", self.Frame, "BOTTOMLEFT");
    self.TextFields[4] = Text4;
        
	-- Declare the fifth text field at CENTER
    local Text5 = self.Frame:CreateFontString(nil, "ARTWORK");
    Text5:SetFontObject("NumberFontNormal");
    Text5:SetJustifyH("CENTER");
    Text5:SetPoint("CENTER", self.Frame, "CENTER");
    self.TextFields[5] = Text5;
end

function TextSubs.ButtonMixin:SetDefaults()
	self.InfoTable.TextSubs = {"", "$(key)", "$(itemcount)", "", "$(cd)"};
end

function TextSubs.ButtonMixin:LoadSettings()
	local TextSubs = self.InfoTable.TextSubs;
	self:SetTextField(1, TextSubs[1]);
	self:SetTextField(2, TextSubs[2]);
	self:SetTextField(3, TextSubs[3]);
	self:SetTextField(4, TextSubs[4]);
	self:SetTextField(5, TextSubs[5]);
end

-- Set a textsub field, generate an unformatted string to format it later on
function TextSubs.ButtonMixin:SetTextField(Field, Text)
	-- Make sure field is between 1 & 5 and Text is a string
    if(Field > 5 or Field < 1 or type(Text) ~= "string") then return; end
    local SubInfo = self.TextSubs[Field];
    local PrefixList = SubInfo.PrefixList;
    local SuffixList = SubInfo.SuffixList;
	local SubCache = SubInfo.SubCache;
	-- Clear the current Prefix & Suffix cache
    for Key, Value in pairs(PrefixList) do
        PrefixList[Key] = nil;
        SuffixList[Key] = nil;
		SubCache[Key] = nil;
    end
	
	local ParsedText = Text;
	-- Loop through all sub matches of the $(...) format
	for Match in string.gmatch(Text, "%$%((.-)%)") do
		-- Find out the prefix and the suffix
		local _, _, Prefix, Suffix = string.find(Match, "(%a*) ?(.*)");
		table.insert(PrefixList, Prefix);
		table.insert(SuffixList, Suffix);
		-- Replace what we just found by %s for string.match later on, also make sure only one replacement is done maximum
		ParsedText = string.gsub(ParsedText, "%$%(" .. Match .."%)", "%%%s", 1);
	end
	FlexBar2:Debug("Parsed Text: [", ParsedText, "]");
	-- Save our parsing results
	SubInfo.ParsedText = ParsedText;
	self:UpdateTextSub();
	self:MatchTextField(Field);
end

function TextSubs.ButtonMixin:UpdateTextSub(TextSub, ...)
	-- Loop through all fields
	for Field = 1, 5 do
		-- Boolean local wether to rematch or not
		local PlanifyRematch = false;
		local SubInfo = self.TextSubs[Field];
		-- Loop through all prefixes, continue if the prefix matches the current sub
		for Key, Prefix in ipairs(SubInfo.PrefixList) do
			if(Prefix == TextSub or TextSub == nil) then
				local OldValue = SubInfo.SubCache[Key];
				local Suffix = SubInfo.SuffixList[Key];
				-- Sub replacement is a function, set its return value
				if(type(TextSubList[TextSub]) == "function") then
					SubInfo.SubCache[Key] = TextSubList[TextSub](self, Suffix, ...) or "";
				-- Sub replacement is a string, just set its value
				elseif(type(TextSubList[TextSub]) == "function") then
					SubInfo.SubCache[Key] = TextSubList[TextSub];
				else
					SubInfo.SubCache[Key] = "";
				end
				-- If a rematch is currently not planned and if the Old value is not equal to the new one, planify one
				if(PlanifyRematch == false and OldValue ~= SubInfo.SubCache[Key]) then PlanifyRematch = true; end
			end
		end
		if(PlanifyRematch) then self:MatchTextField(Field); end		
	end
end

function TextSubs.ButtonMixin:MatchTextField(Field)
	-- Anti-idiot check
	if(Field > 5 or Field < 1) then return; end
	local SubInfo = self.TextSubs[Field];
	-- Format the string with the cached subs
	FlexBar2:Debug("Updating field ", Field, " on button ", self.Name);
	self.TextFields[Field]:SetText(SubInfo.ParsedText:format(unpack(SubInfo.SubCache)));
end

--[[function Layout.ButtonWaterfall(Action, ButtonName, ButtonTable)
	if(Action == "edit") then
		 Waterfall:AddControl(
		"type", "label",
		"text", "Button Scale: ");
        Waterfall:AddControl(
		"type", "textbox",
		"width", "50",
		"noNewLine", true,
		"causeRefresh", false,		
		"changedFunc", function(Scale) TempScale = Scale; end,
		"getFunc", function() return FlexBar2:GetSetting(ButtonTable, "Scale") or 1 end);
	elseif(Action == "save") then
        local ButtonList = FlexBar2.Buttons;
		
	end
end--]]
